@@ -15,6 +15,9 @@ use Fcntl;    # O_RDONLY, etc.
 
 use constant SUPPORTED_SYSOPEN_MODES => O_RDONLY | O_WRONLY | O_RDWR | O_APPEND | O_TRUNC | O_EXCL | O_CREAT | O_NOFOLLOW;
 
+use constant BROKEN_SYMLINK   => bless {}, "A::BROKEN::SYMLINK";
+use constant CIRCULAR_SYMLINK => bless {}, "A::CIRCULAR::SYMLINK";
+
 # we're going to use carp but the errors should come from outside of our package.
 use Carp ();
 $Carp::Internal{__PACKAGE__}++;
@@ -40,11 +43,11 @@ Test::MockFile - Allows tests to validate code that can interact with files with
 
 =head1 VERSION
 
-Version 0.013
+Version 0.014
 
 =cut
 
-our $VERSION = '0.013';
+our $VERSION = '0.014';
 
 our %files_being_mocked;
 
@@ -196,7 +199,7 @@ sub file {
     my ( $class, $file, $contents, @stats ) = @_;
 
     ( defined $file && length $file ) or die("No file provided to instantiate $class");
-    $files_being_mocked{$file} and die("It looks like $file is already being mocked. We don't support double mocking yet.");
+    _get_file_object($file) and die("It looks like $file is already being mocked. We don't support double mocking yet.");
 
     my %stats;
     if ( scalar @stats == 1 ) {
@@ -270,7 +273,7 @@ sub symlink {
     ( defined $file && length $file ) or die("No file provided to instantiate $class");
     ( !defined $readlink || length $readlink ) or die("No file provided for $file to point to in $class");
 
-    $files_being_mocked{$file} and die("It looks like $file is already being mocked. We don't support double mocking yet.");
+    _get_file_object($file) and die("It looks like $file is already being mocked. We don't support double mocking yet.");
 
     return $class->new(
         {
@@ -299,7 +302,7 @@ sub dir {
     my ( $class, $dir_name, $contents, @stats ) = @_;
 
     ( defined $dir_name && length $dir_name ) or die("No directory name provided to instantiate $class");
-    $files_being_mocked{$dir_name} and die("It looks like $dir_name is already being mocked. We don't support double mocking yet.");
+    _get_file_object($dir_name) and die("It looks like $dir_name is already being mocked. We don't support double mocking yet.");
 
     # Because undef means it's a missing dir.
     if ( defined $contents ) {
@@ -442,31 +445,30 @@ sub _mock_stat {
       : $type eq 'lstat' ? 0
       :                    die("Unexpected stat type '$type'");
 
+    # Overload::FileCheck should always send 2 args.
     if ( scalar @_ != 2 ) {
         _real_file_access_hook( $type, [$file_or_fh] );
         return FALLBACK_TO_REAL_OP();
     }
 
+    # Overload::FileCheck should always send something and be handling undef on its own??
     if ( !defined $file_or_fh || !length $file_or_fh ) {
         _real_file_access_hook( $type, [$file_or_fh] );
         return FALLBACK_TO_REAL_OP();
     }
 
+    # Find the path, following the symlink if required.
     my $file = _find_file_or_fh( $file_or_fh, $follow_link );
 
-    # If it's a broken link, we don't want to fall back, we want to return an empty array.
-    if ( $follow_link && !$file && _find_file_or_fh( $file_or_fh, 0 ) ) {
-        return [];
-    }
-
-    return $file if ref $file eq 'ARRAY';    # Allow an ELOOP to fall through here.
+    return [] if $file && $file eq BROKEN_SYMLINK;      # Allow an ELOOP to fall through here.
+    return [] if $file && $file eq CIRCULAR_SYMLINK;    # Allow an ELOOP to fall through here.
 
     if ( !defined $file or !length $file ) {
         _real_file_access_hook( $type, [$file_or_fh] );
         return FALLBACK_TO_REAL_OP();
     }
 
-    my $file_data = $files_being_mocked{$file};
+    my $file_data = _get_file_object($file);
     if ( !$file_data ) {
         _real_file_access_hook( $type, [$file_or_fh] );
         return FALLBACK_TO_REAL_OP();
@@ -487,50 +489,56 @@ sub _get_file_object {
     return $files_being_mocked{$file};
 }
 
+# This subroutine finds the absolute path to a file, returning the absolute path of what it ultimately points to.
+# If it is a broken link or what was passed in is undef or '', then we return undef.
+
 sub _find_file_or_fh {
-    my ( $file_or_fh, $follow_link, $depth, $parent ) = @_;
+    my ( $file_or_fh, $follow_link, $depth ) = @_;
 
-    my $file = _fh_to_file($file_or_fh);
-    $file_or_fh = $file if $file;
-    my $mock_object = $files_being_mocked{$file_or_fh};
+    # Find the file handle or fall back to just using the abs path of $file_or_fh
+    my $absolute_path_to_file = _fh_to_file($file_or_fh) // _abs_path_to_file($file_or_fh) // '';
 
-    if ( $parent and !$mock_object ) {
-        return;
-    }
+    # Get the pointer to the object.
+    my $mock_object = $files_being_mocked{$absolute_path_to_file};
 
-    return $file_or_fh unless $follow_link && $mock_object && $mock_object->is_link;
+    # If we're following a symlink and the path we came to is a dead end (broken symlink), then return BROKEN_SYMLINK up the stack.
+    return BROKEN_SYMLINK if $depth and !$mock_object;
 
-    if ( !$mock_object ) {
-        return [] if $depth;
-        return $file;
-    }
+    # If the link we followed isn't a symlink, then return it.
+    return $absolute_path_to_file unless $mock_object && $mock_object->is_link;
 
-    return $file unless $mock_object->is_link;
+    # ##############
+    # From here on down we're only dealing with symlinks.
+    # ##############
 
-    $depth ||= 0;
+    # If we weren't told to follow the symlink then SUCCESS!
+    return $absolute_path_to_file unless $follow_link;
+
+    # This is still a symlink keep going. Bump our depth counter.
     $depth++;
 
-    #Protect against circular loops.
+    #Protect against circular symlink loops.
     if ( $depth > FOLLOW_LINK_MAX_DEPTH ) {
         $! = ELOOP;
-        return [];
+        return CIRCULAR_SYMLINK;
     }
 
-    return _find_file_or_fh( $mock_object->readlink, 1, $depth, $file_or_fh );
+    return _find_file_or_fh( $mock_object->readlink, 1, $depth );
 }
 
+# Tries to find $fh as a open file handle in one of the mocked files.
+
 sub _fh_to_file {
     my ($fh) = @_;
 
-    return unless defined $fh;
-    return unless length $fh;
+    return unless defined $fh && length $fh;
 
     # See if $fh is a file handle. It might be a path.
-    foreach my $file_name ( keys %files_being_mocked ) {
+    foreach my $file_name ( sort keys %files_being_mocked ) {
         my $mock_fh = $files_being_mocked{$file_name}->{'fh'};
 
-        next unless $mock_fh;              # File isn't open.
-        next unless "$mock_fh" eq "$fh";
+        next unless $mock_fh;               # File isn't open.
+        next unless "$mock_fh" eq "$fh";    # This mock doesn't have this file handle open.
 
         return $file_name;
     }
@@ -1000,8 +1008,8 @@ sub _goto_is_available {
 
 BEGIN {
     *CORE::GLOBAL::open = sub(*;$@) {
-        my $abs_path = _abs_path_to_file( $_[2] );
 
+        # We're not supporting 2 arg or 1 arg opens yet.
         # open(my $fh, ">filehere"); # Just don't do this. It's bad.
         if ( scalar @_ != 3 ) {
             _real_file_access_hook( "open", \@_ );
@@ -1017,18 +1025,23 @@ BEGIN {
             }
         }
 
+        my $abs_path = _find_file_or_fh( $_[2], 1 );    # Follow the link.
+        die if ( !$abs_path );
+        die if $abs_path eq BROKEN_SYMLINK;
+        my $mock_file = _get_file_object($abs_path);
+
         my $mode = $_[1];
 
         # For now we're going to just strip off the binmode and hope for the best.
         $mode =~ s/(:.+$)//;
         my $encoding_mode = $1;
 
-        # TODO: We technically need to support this.
-        # open(my $fh, "-|", "/bin/hostname"); # Read from command
-        # open(my $fh, "|-", "/bin/passwd"); # Write to command
+        # TODO: We don't yet support |- or -|
+        # TODO: We don't yet support modes outside of > < >> +< +> +>>
+        # We just pass through to open if we're not mocking the file right now.
         if (   ( $mode eq '|-' || $mode eq '-|' )
             or !grep { $_ eq $mode } qw/> < >> +< +> +>>/
-            or !defined $files_being_mocked{$abs_path} ) {
+            or !defined $mock_file ) {
             _real_file_access_hook( "open", \@_ );
             goto \&CORE::open if _goto_is_available();
             if ( @_ == 1 ) {
@@ -1042,9 +1055,7 @@ BEGIN {
             }
         }
 
-        #
-        my $followed_link = _find_file_or_fh( $abs_path, 1 );    # Follow the link.
-        my $mock_file = _get_file_object($followed_link);
+        # At this point we're mocking the file. Let's do it!
 
         # If contents is undef, we act like the file isn't there.
         if ( !defined $mock_file->{'contents'} && grep { $mode eq $_ } qw/< +</ ) {
@@ -1057,20 +1068,20 @@ BEGIN {
         $rw .= 'w' if grep { $_ eq $mode } qw/+< +> +>> > >>/;
 
         $_[0] = IO::File->new;
-        tie *{ $_[0] }, 'Test::MockFile::FileHandle', $followed_link, $rw;
+        tie *{ $_[0] }, 'Test::MockFile::FileHandle', $abs_path, $rw;
 
         # This is how we tell if the file is open by something.
 
-        $files_being_mocked{$abs_path}->{'fh'} = $_[0];
+        $mock_file->{'fh'} = $_[0];
         Scalar::Util::weaken( $_[0] );    # Will this make it go out of scope?
 
         # Fix tell based on open options.
         if ( $mode eq '>>' or $mode eq '+>>' ) {
-            $files_being_mocked{$abs_path}->{'contents'} //= '';
-            seek $_[0], length( $files_being_mocked{$abs_path}->{'contents'} ), 0;
+            $mock_file->{'contents'} //= '';
+            seek $_[0], length( $mock_file->{'contents'} ), 0;
         }
         elsif ( $mode eq '>' or $mode eq '+>' ) {
-            $files_being_mocked{$abs_path}->{'contents'} = '';
+            $mock_file->{'contents'} = '';
         }
 
         return 1;
@@ -1090,15 +1101,14 @@ BEGIN {
     # 8 - O_NOFOLLOW - Fail if the last path component is a symbolic link.
 
     *CORE::GLOBAL::sysopen = sub(*$$;$) {
-        my $abs_path = _abs_path_to_file( $_[1] );
+        my $mock_file = _get_file_object( $_[1] );
 
-        if ( !defined $files_being_mocked{$abs_path} ) {
+        if ( !$mock_file ) {
             _real_file_access_hook( "sysopen", \@_ );
             goto \&CORE::sysopen if _goto_is_available();
             return CORE::sysopen( $_[0], $_[1], @_[ 2 .. $#_ ] );
         }
 
-        my $mock_file    = $files_being_mocked{$abs_path};
         my $sysopen_mode = $_[2];
 
         # Not supported by my linux vendor: O_EXLOCK | O_SHLOCK
@@ -1142,6 +1152,8 @@ BEGIN {
             return;
         }
 
+        my $abs_path = $mock_file->{'file_name'};
+
         $_[0] = IO::File->new;
         tie *{ $_[0] }, 'Test::MockFile::FileHandle', $abs_path, $rw;
 
@@ -1163,8 +1175,7 @@ BEGIN {
     };
 
     *CORE::GLOBAL::opendir = sub(*$) {
-
-        my $abs_path = _abs_path_to_file( $_[1] );
+        my $mock_dir = _get_file_object( $_[1] );
 
         # 1 arg Opendir doesn't work??
         if ( scalar @_ != 2 or !defined $_[1] ) {
@@ -1175,15 +1186,13 @@ BEGIN {
             return CORE::opendir( $_[0], @_[ 1 .. $#_ ] );
         }
 
-        if ( !defined $files_being_mocked{$abs_path} ) {
+        if ( !$mock_dir ) {
             _real_file_access_hook( "opendir", \@_ );
             print "Real open\n";
             goto \&CORE::opendir if _goto_is_available();
             return CORE::opendir( $_[0], $_[1] );
         }
 
-        my $mock_dir = $files_being_mocked{$abs_path};
-
         if ( !defined $mock_dir->{'contents'} ) {
             $! = ENOENT;
             return undef;
@@ -1198,6 +1207,7 @@ BEGIN {
         }
 
         # This is how we tell if the file is open by something.
+        my $abs_path = $mock_dir->{'file_name'};
         $mock_dir->{'obj'} = Test::MockFile::DirHandle->new( $abs_path, $mock_dir->{'contents'} );
         $mock_dir->{'fh'} = "$_[0]";
 
@@ -1358,7 +1368,7 @@ BEGIN {
             return;
         }
 
-        my $mock_object = $files_being_mocked{ _abs_path_to_file($file) };
+        my $mock_object = _get_file_object($file);
         if ( !$mock_object ) {
             goto \&CORE::readlink if _goto_is_available();
             return CORE::readlink($file);
